from typing import List

from qdrant_client import QdrantClient, AsyncQdrantClient
from qdrant_client.models import (
    Distance,
    FieldCondition,
    Filter,
    HnswConfigDiff,
    KeywordIndexParams,
    MatchValue,
    PointStruct,
    ScoredPoint,
    VectorParams,
    HasIdCondition,
)

from best_shot.types import (
    dump_io_value,
    Embedding,
    parse_io_value,
    Shot,
    ShotWithSimilarity,
)

from .base import Store


class QdrantBase(Store):
    collection_name: str

    def __init__(self, collection_name: str):
        self.collection_name = collection_name

    def collection_config(self, size: int, distance: Distance, payload_m: int = 16):
        """
        Use these as kwargs for `.create_collection` for optimal performance.
        """
        return dict(
            collection_name=self.collection_name,
            vectors_config=VectorParams(size=size, distance=distance),
            hnsw_config=HnswConfigDiff(payload_m=payload_m, m=0),
        )

    def payload_config(self):
        """
        Use these as kwargs for `.create_payload_index` for optimal performance.
        """
        return dict(
            collection_name=self.collection_name,
            field_name="namespace",
            field_schema=KeywordIndexParams(type="keyword", is_tenant=True),
        )

    def _filter(self, namespace: str, ids: List[str] = []):
        must = [FieldCondition(key="namespace", match=MatchValue(value=namespace))]
        if ids:
            must.append(HasIdCondition(has_id=ids))
        return Filter(must=must)

    def _shots_to_points(
        self,
        shots: List[Shot],
        embeddings: List[Embedding],
        namespace: str,
    ) -> List[PointStruct]:
        return [
            PointStruct(
                id=shot.id,
                vector=embedding,
                payload={
                    "namespace": namespace,
                    "inputs": dump_io_value(shot.inputs),
                    "outputs": dump_io_value(shot.outputs),
                },
            )
            for shot, embedding in zip(shots, embeddings)
        ]

    def _search_to_list(self, results: List[ScoredPoint]) -> List[ShotWithSimilarity]:
        return [
            (
                Shot(
                    id=result.id,
                    inputs=parse_io_value(result.payload["inputs"]),
                    outputs=parse_io_value(result.payload["outputs"]),
                ),
                result.score,
            )
            for result in results
        ]


class QdrantStore(QdrantBase):
    client: QdrantClient

    def __init__(self, client: QdrantClient, collection_name: str):
        super().__init__(collection_name)
        self.client = client

    def add(self, shots: List[Shot], embeddings: List[Embedding], namespace: str):
        self.client.upsert(
            collection_name=self.collection_name,
            points=self._shots_to_points(shots, embeddings, namespace),
        )

    def remove(self, ids: List[str], namespace: str):
        self.client.delete(
            collection_name=self.collection_name,
            points_selector=self._filter(namespace, ids),
        )

    def clear(self, namespace: str):
        self.client.delete(
            collection_name=self.collection_name,
            points_selector=self._filter(namespace),
        )

    def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> List[ShotWithSimilarity]:
        results = self.client.search(
            collection_name=namespace,
            query_vector=embedding,
            query_filter=self._filter(namespace),
            limit=limit,
        )
        return self._search_to_list(results)


class AsyncQdrantStore(QdrantBase):
    client: AsyncQdrantClient

    def __init__(self, client: AsyncQdrantClient, collection_name: str):
        super().__init__(collection_name)
        self.client = client

    async def add(self, shots: List[Shot], embeddings: List[Embedding], namespace: str):
        await self.client.upsert(
            collection_name=namespace,
            points=self._shots_to_points(shots, embeddings, namespace),
        )

    async def remove(self, ids: List[str], namespace: str):
        await self.client.delete(
            collection_name=namespace,
            points_selector=self._filter(namespace, ids),
        )

    async def clear(self, namespace: str):
        await self.client.delete(
            collection_name=namespace,
            points_selector=self._filter(namespace),
        )

    async def list(
        self, embedding: Embedding, namespace: str, limit: int
    ) -> List[ShotWithSimilarity]:
        results = await self.client.search(
            collection_name=namespace,
            query_vector=embedding,
            query_filter=self._filter(namespace),
            limit=limit,
        )
        return self._search_to_list(results)



from dataclasses import dataclass
from typing import overload

from best_shot.types import (
    Shot,
    dump_io_value,
    Datum,
    IO,
    ShotWithSimilarity,
    is_io_value,
)

from .embed.base import Embedder
from .store.base import Store


@dataclass
class BestShots:
    embed: Embedder
    store: Store

    @overload
    def add(
        self,
        inputs: IO,
        outputs: IO,
        *,
        id: str = "",
        namespace: str = "default",
    ) -> str: ...

    @overload
    def add(
        self,
        data: list[Datum],
        *,
        namespace: str = "default",
    ) -> list[str]: ...

    def add(
        self,
        maybe_inputs: IO | list[Datum],
        maybe_outputs: IO | None = None,
        *,
        id: str = "",
        namespace: str = "default",
    ) -> str | list[str]:
        is_io_args = is_io_value(maybe_inputs) and is_io_value(maybe_outputs)
        data: list[Datum] = (
            [(maybe_inputs, maybe_outputs, id)] if is_io_args else maybe_inputs
        )
        shots = [Shot(*datum) for datum in data]
        embeddings = self.embed([shot.key for shot in shots])
        self.store.add(shots, embeddings, namespace)

        ids = [shot.id for shot in shots]
        return ids[0] if is_io_args else ids

    @overload
    def remove(
        self,
        ids: list[str],
        *,
        namespace: str = "default",
    ): ...

    @overload
    def remove(
        self,
        inputs: dict,
        outputs: dict,
        *,
        id: str = "",
        namespace: str = "default",
    ): ...

    @overload
    def remove(
        self,
        data: list[Datum],
        *,
        namespace: str = "default",
    ): ...

    def remove(
        self,
        maybe_inputs: IO | list[Datum],
        maybe_outputs: IO | None = None,
        *,
        id: str = "",
        namespace: str = "default",
    ):
        is_io_args = is_io_value(maybe_inputs) and is_io_value(maybe_outputs)
        data: list[Datum] = (
            [(maybe_inputs, maybe_outputs, id)] if is_io_args else maybe_inputs
        )
        ids = data if isinstance(data[0], str) else [Shot(*datum).id for datum in data]
        self.store.remove(ids, namespace)

    def clear(self, namespace: str = "default"):
        self.store.clear(namespace)

    def list(
        self,
        inputs: IO,
        *,
        namespace: str = "default",
        limit: int = 5,
    ) -> list[ShotWithSimilarity]:
        [embedding] = self.embed([dump_io_value(inputs)])
        return self.store.list(embedding, namespace, limit)



from weaviate.classes.data import DataObject
from weaviate.classes.query import Filter
from weaviate.collections import Collection, CollectionAsync
from weaviate.collections.classes.internal import QuerySearchReturnType

from best_shot.types import (
    dump_io_value,
    Embedding,
    parse_io_value,
    Shot,
    ShotWithSimilarity,
)

from .base import Store


class WeaviateBase(Store):
    @staticmethod
    def _shots_to_data_objects(
        shots: list[Shot],
        embeddings: list[Embedding],
        namespace: str,
    ) -> list[DataObject]:
        return [
            DataObject(
                uuid=shot.id,
                vector=embedding,
                properties={
                    "namespace": namespace,
                    "inputs": dump_io_value(shot.inputs),
                    "outputs": dump_io_value(shot.outputs),
                },
            )
            for shot, embedding in zip(shots, embeddings)
        ]

    @staticmethod
    def _response_to_shots(response: QuerySearchReturnType) -> list[ShotWithSimilarity]:
        return [
            (
                Shot(
                    parse_io_value(o.properties["inputs"]),
                    parse_io_value(o.properties["outputs"]),
                    str(o.uuid),
                ),
                o.metadata.distance,
            )
            for o in response.objects
        ]

    @staticmethod
    def _namespace_filter(namespace: str) -> Filter:
        return Filter.by_property("namespace").equal(namespace)


class WeaviateStore(WeaviateBase):
    collection: Collection

    def __init__(self, collection: Collection):
        self.collection = collection

    def add(
        self,
        shots: list[Shot],
        embeddings: list[Embedding],
        namespace: str,
    ):
        self.collection.data.insert_many(
            self._shots_to_data_objects(shots, embeddings, namespace)
        )

    def remove(self, ids: list[str], namespace: str):
        self.collection.data.delete_many(
            self._namespace_filter(namespace) & Filter.by_id().contains_any(ids)
        )

    def clear(self, namespace: str):
        self.collection.data.delete_many(self._namespace_filter(namespace))

    def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> list[ShotWithSimilarity]:
        response = self.collection.query.near_vector(
            embedding,
            filters=self._namespace_filter(namespace),
            limit=limit,
        )

        return self._response_to_shots(response)


class AsyncWeaviateStore(WeaviateBase):
    collection: CollectionAsync

    def __init__(self, collection: CollectionAsync):
        self.collection = collection

    async def add(
        self,
        shots: list[Shot],
        embeddings: list[Embedding],
        namespace: str,
    ):
        await self.collection.data.insert_many(
            self._shots_to_data_objects(shots, embeddings, namespace)
        )

    async def remove(self, ids: list[str], namespace: str):
        await self.collection.data.delete_many(
            self._namespace_filter(namespace) & Filter.by_id().contains_any(ids)
        )

    async def clear(self, namespace: str):
        await self.collection.data.delete_many(self._namespace_filter(namespace))

    async def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> list[ShotWithSimilarity]:
        response = await self.collection.query.near_vector(
            embedding,
            filters=self._namespace_filter(namespace),
            limit=limit,
        )

        return self._response_to_shots(response)



from itertools import chain
from typing import Iterable

import ujson

from best_shot.types import IO, Shot


def to_str(value: IO) -> str:
    if isinstance(value, dict):
        return ujson.dumps(value)
    return str(value)


def flatten(iterable: Iterable[Iterable]) -> list:
    return list(chain.from_iterable(iterable))


def shots_to_messages(shots: list[Shot]) -> list[dict]:
    return flatten(
        [
            [
                {"role": "user", "content": to_str(shot.inputs)},
                {"role": "assistant", "content": to_str(shot.outputs)},
            ]
            for shot in shots
        ]
    )



from best_shot.types import Shot, dump_io_value, id_io_value, is_io_value


def test_is_io_value():
    assert is_io_value("test") is True
    assert is_io_value({"key": "value"}) is True
    assert is_io_value(123) is False
    assert is_io_value(["test"]) is False


def test_data_key():
    data = {"b": 2, "a": 1}
    assert dump_io_value(data) == '{"a":1,"b":2}'


def test_data_hash():
    data = {"test": "value"}
    expected_hash = "f98be16ebfa861cb39a61faff9e52b33f5bcc16bb6ae72e728d226dc07093932"
    assert id_io_value(data) == expected_hash


def test_init_with_id():
    shot = Shot({"input": "test"}, {"output": "result"}, "test_id")
    assert shot.inputs == {"input": "test"}
    assert shot.outputs == {"output": "result"}
    assert shot.id == "test_id"


def test_init_without_id():
    inputs = {"input": "test"}
    shot = Shot(inputs, {"output": "result"})
    assert shot.id == id_io_value(inputs)


def test_key_property():
    inputs = {"input": "test"}
    shot = Shot(inputs, {"output": "result"})
    assert shot.key == dump_io_value(inputs)



from qdrant_client.local.qdrant_local import QdrantLocal
import pytest

from best_shot.store.qdrant import QdrantStore, Shot, Distance
from best_shot.types import id_io_value


@pytest.fixture
def client():
    return QdrantLocal(":memory:")


@pytest.fixture
def store(client: QdrantLocal):
    s = QdrantStore(client=client, collection_name="test")
    client.create_collection(**s.collection_config(2, Distance.COSINE))
    client.create_payload_index(**s.payload_config())
    yield s
    client.delete_payload_index("test", "namespace")
    client.delete_collection("test")


def test_crud(store: QdrantStore):
    shots = [
        Shot("input1", "output1", id_io_value("id1")),
        Shot("input2", "output2", id_io_value("id2")),
    ]
    embeddings = [[1.0, 2.0], [3.0, 4.0]]
    namespace = "test"

    store.add(shots, embeddings, namespace)
    store.add(shots, embeddings, namespace)

    results = store.list([1.0, 2.0], namespace, 2)
    assert len(results) == 2
    assert results[0][0].id == id_io_value("id1")
    assert results[1][0].id == id_io_value("id2")

    store.remove([id_io_value("id1")], namespace)
    results = store.list([1.0, 2.0], namespace, 2)
    assert len(results) == 1

    store.clear(namespace)
    results = store.list([1.0, 2.0], namespace, 2)
    assert len(results) == 0


def test_structured_io(store: QdrantStore):
    shots = [
        Shot({"key": "input1"}, {"key": "output1"}, id_io_value("id1")),
        Shot({"key": "input2"}, {"key": "output2"}, id_io_value("id2")),
    ]
    embeddings = [[1.0, 2.0], [3.0, 4.0]]
    namespace = "test"

    store.add(shots, embeddings, namespace)

    (shot_0, _distance_0), (shot_1, _distance_1) = store.list([1.0, 2.0], namespace, 2)

    assert shot_0 == shots[0]
    assert shot_1 == shots[1]



from dataclasses import dataclass

from sentence_transformers import SentenceTransformer

from best_shot.types import Embedding
from best_shot.utils.asyncio import asyncify_class
from .base import AsyncEmbedder, Embedder


@dataclass
class TransformersEmbedder(Embedder):
    model: SentenceTransformer

    def __call__(self, inputs: list[str]) -> list[Embedding]:
        return self.model.encode(inputs).tolist()


@asyncify_class
class AsyncTransformersEmbedder(TransformersEmbedder, AsyncEmbedder): ...



from .async_client import AsyncBestShots
from .client import BestShots
from .types import Shot, Embedding, ShotWithSimilarity, IO, Datum
from .utils.format import shots_to_messages

__all__ = [
    "AsyncBestShots",
    "BestShots",
    "Datum",
    "Embedding",
    "IO",
    "Shot",
    "shots_to_messages",
    "ShotWithSimilarity",
]



from typing import Any, Awaitable, Callable

from best_shot.types import Embedding
from .base import AsyncEmbedder, Embedder


class LiteLLMEmbedder(Embedder):
    def __init__(self, embedder: Callable[[str], Any]):
        self.embedder = embedder

    def __call__(self, inputs: list[str]) -> list[Embedding]:
        response = self.embedder(inputs)
        return [r["embedding"] for r in response["data"]]


class AsyncLiteLLMEmbedder(AsyncEmbedder):
    def __init__(self, embedder: Callable[[str], Awaitable[Any]]):
        self.embedder = embedder

    async def __call__(self, inputs: list[str]) -> list[Embedding]:
        response = await self.embedder(inputs)
        return [r["embedding"] for r in response["data"]]



from abc import abstractmethod

from best_shot.types import Embedding


class Embedder:
    @abstractmethod
    def __call__(self, inputs: list[str]) -> list[Embedding]: ...


class AsyncEmbedder(Embedder):
    @abstractmethod
    async def __call__(self, inputs: list[str]) -> list[Embedding]: ...



from abc import abstractmethod

from best_shot.types import Embedding, Shot, ShotWithSimilarity


class Store:
    @abstractmethod
    def add(
        self,
        shots: list[Shot],
        embeddings: list[Embedding],
        namespace: str,
    ): ...

    @abstractmethod
    def remove(
        self,
        ids: list[str],
        namespace: str,
    ): ...

    @abstractmethod
    def clear(self, namespace: str): ...

    @abstractmethod
    def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> list[ShotWithSimilarity]: ...


class AsyncStore(Store):
    @abstractmethod
    async def add(
        self,
        shots: list[Shot],
        embeddings: list[Embedding],
        namespace: str,
    ): ...

    @abstractmethod
    async def remove(self, ids: list[str], namespace: str): ...

    @abstractmethod
    async def clear(self, namespace: str): ...

    @abstractmethod
    async def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> list[ShotWithSimilarity]: ...



import pytest
from sentence_transformers import SentenceTransformer

from best_shot.client import BestShots
from best_shot.embed.transformers import TransformersEmbedder
from best_shot.store.memory import MemoryStore
from best_shot.types import Shot


@pytest.fixture(scope="function")
def client():
    return BestShots(
        embed=TransformersEmbedder(model=SentenceTransformer("all-MiniLM-L6-v2")),
        store=MemoryStore(),
    )


def test_functional_flow(client: BestShots):
    inputs = {"a": 1}
    outputs = {"b": 2}

    id = client.add(inputs, outputs)
    assert id == Shot(inputs, outputs).id

    results = client.list(inputs, limit=1)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    results = client.list(inputs, limit=5)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    client.remove(inputs, outputs)
    assert [] == client.list(inputs, limit=1)

    client.add(inputs, outputs)
    client.clear()
    assert [] == client.list(inputs)


def test_dispatch(client: BestShots):
    inputs = {"a": 1}
    outputs = {"b": 2}

    id = client.add(inputs, outputs)
    assert id == Shot(inputs, outputs).id

    client.remove([id])
    assert [] == client.list(inputs)

    [id] = client.add([(inputs, outputs)])
    client.remove([id])
    assert [] == client.list(inputs)

    client.add([(inputs, outputs)])
    client.remove([(inputs, outputs)])
    assert [] == client.list(inputs)

    [id] = client.add([(inputs, outputs, "id")])
    assert id == "id"

    client.remove([(inputs, outputs, "id")])
    assert [] == client.list(inputs)

    client.add([(inputs, outputs, "id")])
    client.remove(["id"])
    assert [] == client.list(inputs)


def test_string_flow(client: BestShots):
    inputs = "User question..."
    outputs = "AI answer..."

    id = client.add(inputs, outputs)
    assert id == Shot(inputs, outputs).id

    results = client.list(inputs, limit=1)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    results = client.list(inputs, limit=5)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    client.remove(inputs, outputs)
    assert [] == client.list(inputs, limit=1)

    client.add(inputs, outputs)
    client.clear()
    assert [] == client.list(inputs)



from typing import List

from chromadb import Collection
from chromadb.api.async_client import AsyncCollection

from best_shot.types import (
    dump_io_value,
    Embedding,
    parse_io_value,
    Shot,
    ShotWithSimilarity,
)

from .base import Store


class ChromaBase(Store):
    @staticmethod
    def _query_to_list(results: dict) -> list[ShotWithSimilarity]:
        return [
            (
                Shot(parse_io_value(inputs), parse_io_value(metadata["outputs"]), id),
                distance,
            )
            for (id, inputs, distance, metadata) in zip(
                results["ids"][0],
                results["documents"][0],
                results["distances"][0],
                results["metadatas"][0],
            )
        ]


class ChromaStore(ChromaBase):
    collection: Collection

    def __init__(self, collection: Collection):
        self.collection = collection

    def add(self, shots: list[Shot], embeddings: list[Embedding], namespace: str):
        self.collection.upsert(
            ids=[shot.id for shot in shots],
            documents=[shot.key for shot in shots],
            embeddings=embeddings,
            metadatas=[
                {"namespace": namespace, "outputs": dump_io_value(s.outputs)}
                for s in shots
            ],
        )

    def remove(self, ids: list[str], namespace: str):
        self.collection.delete(ids=ids, where={"namespace": namespace})

    def clear(self, namespace: str):
        self.collection.delete(where={"namespace": namespace})

    def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> list[ShotWithSimilarity]:
        results = self.collection.query(
            query_embeddings=[embedding],
            n_results=limit,
            where={"namespace": namespace},
        )
        return self._query_to_list(results)


class AsyncChromaStore(ChromaBase):
    collection: AsyncCollection

    async def add(self, shots: list[Shot], embeddings: list[Embedding], namespace: str):
        await self.collection.add(
            ids=[shot.id for shot in shots],
            embeddings=embeddings,
            documents=[shot.key for shot in shots],
            metadatas=[
                {"namespace": namespace, "outputs": dump_io_value(s.outputs)}
                for s in shots
            ],
        )

    async def remove(self, ids: list[str], namespace: str):
        await self.collection.delete(ids=ids, where={"namespace": namespace})

    async def clear(self, namespace: str):
        await self.collection.delete(where={"namespace": namespace})

    async def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> list[ShotWithSimilarity]:
        results = await self.collection.query(
            query_embeddings=[embedding],
            n_results=limit,
            where={"namespace": namespace},
        )
        return self._query_to_list(results)



from collections import defaultdict
from operator import itemgetter

import numpy as np

from best_shot.utils.asyncio import asyncify_class

from .base import Embedding, Shot, ShotWithSimilarity, Store


def cosine_similarity(a: Embedding, b: Embedding) -> float:
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))


class MemoryStore(Store):
    # namespace => id => (shot, embedding)
    _storage: dict[str, dict[str, tuple[Shot, Embedding]]]

    def __init__(self):
        self._storage = defaultdict(dict)

    def add(self, shots: list[Shot], embeddings: list[Embedding], namespace: str):
        for shot, embedding in zip(shots, embeddings):
            self._storage[namespace][shot.id] = (shot, embedding)

    def remove(self, ids: list[str], namespace: str):
        for id in ids:
            del self._storage[namespace][id]

    def clear(self, namespace: str):
        self._storage[namespace].clear()

    def list(
        self,
        embedding: Embedding,
        namespace: str,
        limit: int,
    ) -> list[ShotWithSimilarity]:
        shots_with_similarities = [
            (shot, 1 - cosine_similarity(embedding, emb))
            for (shot, emb) in self._storage[namespace].values()
        ]
        return sorted(shots_with_similarities, key=itemgetter(1), reverse=True)[:limit]


@asyncify_class
class AsyncMemoryStore(MemoryStore): ...



import pytest
from sentence_transformers import SentenceTransformer

from best_shot.async_client import AsyncBestShots
from best_shot.embed.transformers import AsyncTransformersEmbedder
from best_shot.store.memory import AsyncMemoryStore
from best_shot.types import Shot


@pytest.fixture(scope="function")
def client():
    return AsyncBestShots(
        embed=AsyncTransformersEmbedder(model=SentenceTransformer("all-MiniLM-L6-v2")),
        store=AsyncMemoryStore(),
    )


@pytest.mark.anyio
async def test_functional_flow(client: AsyncBestShots):
    inputs = {"a": 1}
    outputs = {"b": 2}

    id = await client.add(inputs, outputs)
    assert id == Shot(inputs, outputs).id

    results = await client.list(inputs, limit=1)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    results = await client.list(inputs, limit=5)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    await client.remove(inputs, outputs)
    assert [] == await client.list(inputs, limit=1)

    await client.add(inputs, outputs)
    await client.clear()
    assert [] == await client.list(inputs)


@pytest.mark.anyio
async def test_dispatch(client: AsyncBestShots):
    inputs = {"a": 1}
    outputs = {"b": 2}

    id = await client.add(inputs, outputs)
    assert id == Shot(inputs, outputs).id

    await client.remove([id])
    assert [] == await client.list(inputs)

    [id] = await client.add([(inputs, outputs)])
    await client.remove([id])
    assert [] == await client.list(inputs)

    await client.add([(inputs, outputs)])
    await client.remove([(inputs, outputs)])
    assert [] == await client.list(inputs)

    [id] = await client.add([(inputs, outputs, "id")])
    assert id == "id"

    await client.remove([(inputs, outputs, "id")])
    assert [] == await client.list(inputs)

    await client.add([(inputs, outputs, "id")])
    await client.remove(["id"])
    assert [] == await client.list(inputs)


@pytest.mark.anyio
async def test_string_flow(client: AsyncBestShots):
    inputs = "User question..."
    outputs = "AI answer..."

    id = await client.add(inputs, outputs)
    assert id == Shot(inputs, outputs).id

    results = await client.list(inputs, limit=1)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    results = await client.list(inputs, limit=5)
    results = [shot for shot, _ in results]
    assert results == [Shot(inputs, outputs)]

    await client.remove(inputs, outputs)
    assert [] == await client.list(inputs, limit=1)

    await client.add(inputs, outputs)
    await client.clear()
    assert [] == await client.list(inputs)



import pytest

from chromadb import Client
from chromadb.api.client import ClientAPI

from best_shot.store.chroma import ChromaStore, Collection, Shot


@pytest.fixture
def client():
    return Client()


@pytest.fixture
def collection(client: ClientAPI):
    return client.create_collection("test", get_or_create=True)


@pytest.fixture
def store(collection: Collection):
    return ChromaStore(collection)


def test_crud(store: ChromaStore, collection: Collection):
    shots = [Shot("input1", "output1", "id1"), Shot("input2", "output2", "id2")]
    embeddings = [[1.0, 2.0], [3.0, 4.0]]
    namespace = "test"

    store.add(shots, embeddings, namespace)
    assert collection.count() == 2

    store.add(shots, embeddings, namespace)
    assert collection.count() == 2

    results = store.list([1.0, 2.0], namespace, 2)
    assert len(results) == 2
    assert results[0][0].id == "id1"
    assert results[1][0].id == "id2"

    store.remove(["id1"], namespace)
    assert collection.count() == 1

    store.clear(namespace)
    assert collection.count() == 0


def test_structured_io(store: ChromaStore):
    shots = [
        Shot({"key": "input1"}, {"key": "output1"}, "id1"),
        Shot({"key": "input2"}, {"key": "output2"}, "id2"),
    ]
    embeddings = [[1.0, 2.0], [3.0, 4.0]]
    namespace = "test"

    store.add(shots, embeddings, namespace)

    (shot_0, distance_0), (shot_1, distance_1) = store.list([1.0, 2.0], namespace, 2)

    assert shot_0 == shots[0]
    assert shot_1 == shots[1]
    assert distance_0 <= distance_1
